from abc import ABC, abstractmethod
from loguru import logger

from .utils import xor


class EPP(ABC):

    @abstractmethod
    def open(self) -> None:
        pass

    @abstractmethod
    def close(self) -> None:
        pass

    @abstractmethod
    def clear_keys(self) -> None:
        pass

    @abstractmethod
    def write_device_key(self, hex_data: str) -> str:
        pass

    @abstractmethod
    def write_master_key(self, hex_data: str) -> str:
        pass

    @abstractmethod
    def write_pin_key(self, hex_data: str) -> str:
        pass

    @abstractmethod
    def write_mac_key(self, hex_data: str) -> str:
        pass

    @abstractmethod
    def read_pin(self, card_no: str, timeout: int = 60) -> str:
        pass

    @abstractmethod
    def on_key_press(self, cnt: int = 0) -> None:
        pass

    def read_mac(self, hex_data: str) -> str:
        if len(hex_data) != 16:
            raise NotImplementedError()

        mab = hex_data.upper().encode('ascii').hex()
        enc = xor(self.des_encrypt(mab[:16]), mab[16:])
        enc = self.des_encrypt(enc)
        return enc[:8].upper().encode('ascii').hex()

    @abstractmethod
    def des_encrypt(self, hex_data: str) -> str:
        pass

    @abstractmethod
    def des_decrypt(self, hex_data: str) -> str:
        pass

    def __enter__(self):
        self.open()

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.close()
        except Exception as e:
            logger.debug(e)

        return True



from loguru import logger

from .pos import POS, POSAPI
from .card_reader import CardReader
from .epp import EPP


class CCBPos(POS):

    def __init__(self, bcr: CardReader, epp: EPP, api: POSAPI, device: str, merchant: str) -> None:
        super().__init__(bcr, epp, device, merchant)
        self.api = api

    def sign_in(self) -> str:
        resp = self.api.exec({
            "tranCode": "0800",
            "device": self.device,
            "merchant": self.merchant
        })

        with self.epp:
            r = self.epp.write_pin_key(resp['wkey'])
            logger.debug('write_pin_key {}, {}', r, resp['wsum'])

            if r.lower() != resp['wsum']:
                raise RuntimeError('Failed to write PIN key')

            r = self.epp.write_mac_key(resp['mkey'])
            logger.debug('write_mac_key {}, {}', r, resp['msum'])
            if r.lower() != resp['msum']:
                raise RuntimeError('Failed to write MAC key')

        return resp['batch']

    def sign_out(self) -> None:
        self.api.exec({
            "tranCode": "0820",
            "device": self.device
        })

    def pay(self, amount: int) -> dict:
        if amount < 1:
            raise RuntimeError('The amount must be greater than zero')

        bci = self.bcr.read_info(amount, '00')
        logger.debug("bci: {}", bci)

        if not bci:
            raise RuntimeError('Failed to read card information')

        pin = None
        with self.epp:
            pin = self.epp.read_pin(bci.pin_pan)
            logger.debug('pin: {}', pin)

        if not pin:
            raise RuntimeError('Failed to read PIN')

        resp = self.api.exec({
            "tranCode": "02002",
            'amount': amount,
            "accountSN": bci.pan_sn,
            "track2": bci.track2,
            "device": self.device,
            "password": pin,
            "ic55": bci.ic55
        })

        mac = None
        with self.epp:
            mac = self.epp.read_mac(resp['data'])
            logger.debug('mac: {}', mac)

        if not mac:
            raise RuntimeError('Failed to calculate MAC')

        result = self.api.exec({
            "tranCode": "0000",
            "id": resp['id'],
            "mac": mac
        })

        return result

    def write_off(self, amount: int, order_no: str, batch: str, auth_code: str) -> dict:
        if amount < 1:
            raise RuntimeError('The amount must be greater than zero')

        bci = self.bcr.read_info(amount, '00')
        logger.debug("bci: {}", bci)

        if not bci:
            raise RuntimeError('Failed to read card information')

        resp = self.api.exec({
            "tranCode": "04001",
            "device": self.device,
            "accountSN": bci.pan_sn,
            "track2": bci.track2,
            "ic55": bci.ic55,
            'amount': amount,
            'orderNo': order_no,
            'batch': batch,
            'authCode': auth_code,
            'merchant': self.merchant
        })

        mac = None
        with self.epp:
            mac = self.epp.read_mac(resp['data'])
            logger.debug('mac: {}', mac)

        if not mac:
            raise RuntimeError('Failed to calculate MAC')

        result = self.api.exec({
            "tranCode": "0000",
            "id": resp['id'],
            "mac": mac
        })

        return result



import base64
import json
import requests

from abc import ABC, abstractmethod
from loguru import logger

from .card_reader import CardReader
from .epp import EPP

__TOKEN__ = '696939654e404e293635284634453324596a427a574938624e7056532b2c36'


class POSAPI(object):

    def __init__(self, url: str, token: str = None, pwd: str = None):
        super(POSAPI, self).__init__()
        self.url = url
        self.token = token
        self.pwd = pwd

    @property
    def certificate(self):
        if self.token and self.pwd:
            return (self.token, self.pwd)
        else:
            return tuple(bytes.fromhex(__TOKEN__).decode('utf-8').split('$'))

    def exec(self, params: dict) -> dict:
        logger.debug('request: {}', params)
        r = requests.post(self.url, auth=self.certificate,
                          data=params, timeout=15)

        if r.status_code != requests.codes.ok:
            r.raise_for_status()

        json_text = base64.decodebytes(r.text.encode('utf-8')).decode('utf-8')
        result = json.loads(json_text)
        logger.debug('response: {}', result)
        return result


class POS(ABC):

    def __init__(self, bcr: CardReader, epp: EPP, device: str, merchant: str) -> None:
        super().__init__()
        self.bcr = bcr
        self.epp = epp
        self.device = device
        self.merchant = merchant

    @abstractmethod
    def sign_in(self) -> str:
        pass

    @abstractmethod
    def sign_out(self) -> None:
        pass

    @abstractmethod
    def pay(self, amount: int) -> dict:
        pass

    @abstractmethod
    def write_off(self, amount: int, order_no: str, batch: str, auth_code: str) -> dict:
        pass



def xor(a: str, b: str) -> str:
    ba = bytes.fromhex(a)
    bb = bytes.fromhex(b)

    result = b''
    for i in range(len(ba)):
        result += (ba[i] ^ bb[i]).to_bytes(1, 'big')

    return result.hex()



from loguru import logger
from typing import Dict

PDU = Dict[str, str]


def unpack(cmd: str) -> PDU:
    if len(cmd) % 2 != 0 or len(cmd) < 8:
        raise ValueError("Value is invalid")

    result = {}
    result['cla'] = cmd[:2]
    result['ins'] = cmd[2:4]
    result['p1'] = cmd[4:6]
    result['p2'] = cmd[6:8]

    if len(cmd) == 10:  # CLA INS P1 P2 Le
        result['le'] == cmd[8:10]

    if len(cmd) > 10:
        lc = cmd[8:10]
        result['lc'] = lc

        r = cmd[10:]
        if len(r) > int(lc, 16):
            result['le'] = r[-2:]
            result['data'] = r[:-2]
        else:
            result['data'] = r

    return result


def pack(pdu: PDU) -> str:
    _cla = pdu.get('cla')
    if not _cla:
        raise ValueError("The 'cla' can not be null")

    _ins = pdu.get('ins')
    if not _ins:
        raise ValueError("The 'ins' can not be null")

    _p1 = pdu.get('p1')
    if not _p1:
        raise ValueError("The 'p1' can not be null")

    _p2 = pdu.get('p2')
    if not _p2:
        raise ValueError("The 'p2' can not be null")

    _lc = pdu.get('lc', '')
    _data = pdu.get('data', '')
    if (not _lc and _data) or (_lc and not _data):
        raise ValueError(
            "The 'lc' and 'data' must be provided together")

    _le = pdu.get('le', '')

    return _cla + _ins + _p1 + _p2 + _lc + _data + _le



import time

from typing import Tuple
from .bc_reader import BCReader, BCInfo

READERS = Tuple[BCReader]


class CardReader(object):

    def __init__(self, readers: READERS, timeout: int = 30):
        super(CardReader, self).__init__()
        self.timeout = timeout
        self.readers = readers

    def read_info(self, amount: int, trade: str, merchant: str = None) -> BCInfo:
        e_time = int(time.time()) + self.timeout
        while int(time.time()) < e_time:
            for reader in self.readers:
                with reader:
                    reader.detect()
                    reader.prepare()
                    return reader.read_info(amount, trade, merchant)

            time.sleep(0.1)

        raise RuntimeError('Failed to read card information')



import random

from typing import Dict
from datetime import datetime
from loguru import logger

from . import tlv

TLVDICT = Dict[str, bytes]


class PBOC(object):

    def __init__(self, runner):
        super(PBOC, self).__init__()
        self.exec = runner

    def run(self, cmd: str) -> TLVDICT:
        logger.debug('cmd: {}', cmd)
        resp = self.exec(cmd)
        result = tlv.decode(bytes.fromhex(resp), flatten=True)
        logger.debug('result: {}', result)
        return result

    def select(self, name: str, next=False) -> TLVDICT:
        cmd = '00a404' + ('02' if next else '00') + \
            '{:02x}'.format(len(name) // 2) + name + 'ff'
        return self.run(cmd)

    def read_record(self, sfi: int, idx: int) -> TLVDICT:
        ctl = '{:02x}'.format(sfi << 3 ^ 4)
        cmd = '00b2' + '{:02x}'.format(idx) + ctl
        return self.run(cmd)

    def get_aids(self, sfi: int) -> list:
        aids = []
        for i in range(1, 16):
            try:
                result = self.read_record(sfi, i)
                aid = result['4F'].hex()
                if aid not in aids:
                    aids.append(aid)
            except:
                break

        logger.debug('aids: {}', aids)
        return aids

    def gpo(self, pdol: bytes, money: int) -> TLVDICT:
        # PDOL
        # 工行 9f7a 01 9f02 06 5f2a 02
        # 农行         9f02 06 5f2a 02 df69 01
        # 建行 9f7a 01 9f02 06 5f2a 02
        # 中行 9f7a 01 9f02 06 5f2a 02 df69 01
        # 交通 9f7a 01 9f02 06 5f2a 02 df69 01
        # 广发 9f7a 01 9f02 06 5f2a 02 df69 01
        # 中原 9f7a 01 9f02 06 5f2a 02
        # 华夏 9f7a 01 9f02 06 5f2a 02 df69 01
        # 平安 9f7a 01 9f02 06 5f2a 02
        logger.debug('pdol: {}', pdol)
        gpo_cmd = ''
        if b'\x9F\x7A' in pdol:
            gpo_cmd += 'ff'
        if b'\x9F\x02' in pdol:
            gpo_cmd += '{:012x}'.format(money)
        if b'\x5F\x2A' in pdol:
            gpo_cmd += '0156'
        if b'\xDF\x69' in pdol:
            gpo_cmd += '00'

        gpo_len = len(gpo_cmd) // 2

        cmd = '80a80000' + \
            '{:02x}'.format(gpo_len + 2) + '83' + \
            '{:02x}'.format(gpo_len) + gpo_cmd + 'ff'
        return self.run(cmd)

    def read_records(self, afl: bytes) -> TLVDICT:
        result = {}
        i = 0
        while i < len(afl):
            part = afl[i: i + 4]

            sfi = part[0] >> 3
            j = part[1]
            e = part[2]
            while j <= e:
                resp = self.read_record(sfi, j)
                result.update(resp)
                j += 1

            i += 4

        return result

    def get_data(self, tag: str) -> TLVDICT:
        return self.run('80ca' + tag)

    def ac(self, cdol1: bytes, amount: int, trade: str, merchant: str = None) -> TLVDICT:
        cmd: bytes = b''

        d_9f02 = None
        if b'\x9f\x02' in cdol1:
            d_9f02 = amount.to_bytes(6, 'big')
            cmd += d_9f02

        if b'\x9f\x03' in cdol1:
            cmd += (0).to_bytes(6, 'big')

        if b'\x9f\x1a' in cdol1:
            cmd += b'\x01\x56'

        d_95 = None
        if b'\x95' in cdol1:
            d_95 = b'\x00\x00\x80\x00\x00'
            cmd += d_95

        if b'\x5f\x2a' in cdol1:
            cmd += b'\x01\x56'

        d_9a = None
        if b'\x9a' in cdol1:
            d_9a = bytes.fromhex(datetime.now().strftime('%y%m%d'))
            cmd += d_9a

        d_9c = None
        if b'\x9c' in cdol1:
            d_9c = bytes.fromhex(trade)
            cmd += d_9c

        d_9f37 = None
        if b'\x9f\x37' in cdol1:
            d_9f37 = random.randint(1000000000, 2000000000).to_bytes(4, 'big')
            cmd += d_9f37

        if b'\x9f\x21' in cdol1:
            cmd += bytes.fromhex(datetime.now().strftime('%H%M%S'))

        if b'\x9f\x4e' in cdol1:
            if merchant is None:
                cmd += (0).to_bytes(14, 'big')
            else:
                cmd += bytes.fromhex(merchant)

        cmd += (0).to_bytes(52-len(cmd), 'big')
        cmd = b'\x80\xae\x80\x00\x34' + cmd + b'\xff'
        result = self.run(cmd.hex())

        if '80' in result:
            result.update({
                '9F27': result['80'][:1],
                '9F26': result['80'][3:11]
            })
            if len(result['80']) > 11:
                result.update({
                    '9F10': result['80'][11:]
                })

        result.update({
            '9F37': d_9f37,
            '95': d_95,
            '9A': d_9a,
            '9C': d_9c,
            '9F02': d_9f02,
            '5F2A': b'\x01\x56',
            '9F1A': b'\x01\x56',
            '9F03': (0).to_bytes(6, 'big'),
            '9F33': b'\x00\x40\x00'
        })
        logger.debug('ac result: {}', result)
        return result



import re

from abc import ABC, abstractmethod
from loguru import logger

from .pboc import PBOC


class BCInfo(object):

    def __init__(self, pan: str = '', pan_sn: str = '', track2: str = '', ic55: str = ''):
        super(BCInfo, self).__init__()
        self.pan = pan  # 主账号
        self.pan_sn = pan_sn  # 主账号序列号
        self.track2 = track2  # 二磁
        self.ic55 = ic55  # 55 域

    @property
    def pin_pan(self) -> str:
        r = re.sub(r'\D', '', self.pan)
        t = ''
        if len(r) <= 12:
            t = r
        else:
            t = r[-13:-1]

        return t.encode().rjust(16, b'\x00').hex()

    def __repr__(self) -> str:
        return 'pan: %s, pan sn: %s, pin pan: %s, track2: %s, ic55: %s' % (self.pan, self.pan_sn, self.pin_pan, self.track2, self.ic55)


class BCReader(ABC):

    @abstractmethod
    def open(self) -> None:
        pass

    @abstractmethod
    def detect(self) -> None:
        pass

    @abstractmethod
    def prepare(self) -> None:
        pass

    @abstractmethod
    def close(self) -> None:
        pass

    @abstractmethod
    def read_info(self, amount: int, trade: str, merchant: str = None) -> BCInfo:
        pass

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.close()
        except Exception as e:
            logger.debug(e)

        return True


class CPUReader(BCReader):

    @abstractmethod
    def run(self, hex_cmd: str) -> str:
        pass


class RFCPUReader(CPUReader):

    def read_info(self, amount: int, trade: str, merchant: str = None) -> BCInfo:
        raise NotImplementedError()


class SmartReader(CPUReader):

    def read_info(self, amount: int, trade: str, merchant: str = None) -> BCInfo:
        records = {}

        _pboc = PBOC(self.run)

        r = _pboc.select('315041592e5359532e4444463031')
        logger.debug(r)

        sfi = r['88'][0]
        aids = _pboc.get_aids(sfi)

        for i in range(len(aids)):
            r = _pboc.select(aids[i], i > 0)

            pdol = r['9F38']
            r = _pboc.gpo(pdol, amount)

            aip = r['80'][:2]
            afls = r['80'][2:]
            records = _pboc.read_records(afls)

            records['82'] = aip

        d_9f13 = _pboc.get_data('9F13')
        records.update(d_9f13)

        d_9f36 = _pboc.get_data('9F36')
        records.update(d_9f36)

        cdol1 = records['8C']
        r = _pboc.ac(cdol1, amount, trade, merchant)
        records.update(r)

        logger.debug('records: {}', records)

        d_55_list = (
            b'\x9f\x26', b'\x08', records['9F26'],
            b'\x9f\x27', b'\x01', records['9F27'],
            b'\x9f\x10', (len(records['9F10'])).to_bytes(
                1, 'big'), records['9F10'],
            b'\x9f\x37', b'\x04', records['9F37'],
            b'\x9f\x36', b'\x02', records['9F36'],
            b'\x95', b'\x05', records['95'],
            b'\x9a', b'\x03', records['9A'],
            b'\x9c', b'\x01', records['9C'],
            b'\x9f\x02', b'\x06', records['9F02'],
            b'\x5f\x2a', b'\x02', records['5F2A'],
            b'\x82', b'\x02', records['82'],
            b'\x9f\x1a', b'\x02', records['9F1A'],
            b'\x9f\x03', b'\x06', records['9F03'],
            b'\x9f\x33', b'\x03', records['9F33']
        )

        result: BCInfo = BCInfo()
        result.ic55 = b''.join(d_55_list).hex()
        result.track2 = records['57'].hex()
        result.pan = records['5A'].hex()
        result.pan_sn = records['5F34'].hex()
        logger.debug('result: {}', result)
        return result



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['msm_has_base']

package_data = \
{'': ['*']}

install_requires = \
['loguru>=0.6.0,<0.7.0', 'requests>=2.28.1,<3.0.0']

setup_kwargs = {
    'name': 'msm-has-base',
    'version': '0.1.8',
    'description': 'Hardware Access Services Base Foundation',
    'long_description': '',
    'author': '袁首京',
    'author_email': 'yuanshoujing@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.9,<4.0.0',
}


setup(**setup_kwargs)
